/*-
 * Copyright (c) 2005 Daniel Braniss <danny@cs.huji.ac.il>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/*
 | iSCSI
 | $Id: iscsi.c,v 1.26 2005/09/11 12:39:14 danny Exp danny $
 */
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/conf.h>
#include <sys/bus.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/ctype.h>
#include <sys/errno.h>
#include <sys/sysctl.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/socketvar.h>
#include <sys/socket.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/ioccom.h>
#include <sys/queue.h>
#include <sys/kthread.h>
#include <sys/mbuf.h>
#include <sys/sysctl.h>
#include <sys/syslog.h>
#include <vm/uma.h>

#include <dev/iscsi/iscsi.h>
#include <dev/iscsi/iscsivar.h>

static struct isc_softc isc;

MALLOC_DEFINE(M_ISCSI, "iSCSI", "iSCSI driver");

#ifdef ISCSI_DEBUG
int iscsi_debug = ISCSI_DEBUG;
SYSCTL_INT(_debug, OID_AUTO, iscsi, CTLFLAG_RW, &iscsi_debug, 0,
	"iSCSI driver debug flag");
#endif

SYSCTL_NODE(_hw, OID_AUTO, iscsi, CTLFLAG_RD, NULL, "iSCSI Subsystem");

static char isid[6+1] = {
     0x80,
     'D',
     'I',
     'B',
     '0',
     '0',
     0
};
SYSCTL_STRING(_hw_iscsi, OID_AUTO, isid, CTLFLAG_RW, isid, 6+1, "Initiator System ID");
TUNABLE_STR("hw.iscsi.isid", isid, 6+1);

static int	i_create_session(struct cdev *dev, int *ndev);

static int	i_ping(struct cdev *dev);
static int	i_send(struct cdev *dev, caddr_t arg, struct thread *td);
static int	i_recv(struct cdev *dev, caddr_t arg, struct thread *td);

static d_open_t iscsi_open;
static d_close_t iscsi_close;
static d_ioctl_t iscsi_ioctl;
#ifdef ISCSI_DEBUG
static d_read_t iscsi_read;
#endif
static struct cdevsw iscsi_cdevsw = {
     .d_version = D_VERSION,
     .d_open	= iscsi_open,
     .d_close	= iscsi_close,
     .d_ioctl	= iscsi_ioctl,
#ifdef ISCSI_DEBUG
     .d_read	= iscsi_read,
#endif
     .d_name	= "iSCSI",
};

static int
iscsi_open(struct cdev *dev, int flags, int otype, struct thread *td)
{
     debug_called(8);

     debug(3, "dev=%d", minor(dev));

     if(minor(dev) > MAX_SESSIONS) {
	  // should not happen
          return ENODEV;
     }
     if(minor(dev) == MAX_SESSIONS) {
#if 1
	  struct isc_softc *sc = (struct isc_softc *)dev->si_drv1;

	  // this should be in iscsi_start
	  if(sc->cam_sim == NULL)
	       ic_init(sc);
#endif
     }
     return 0;
}

static int
iscsi_close(struct cdev *dev, int flag, int otyp, struct thread *td)
{
     struct isc		*sc;
     isc_session_t	*sp;

     debug_called(8);

     sc = (struct isc *)dev->si_drv1;
     if(minor(dev) == MAX_SESSIONS) {
	  return 0;
     }
     sp = (isc_session_t *)dev->si_drv2;
     if(sp != NULL) {
	  sdebug(2, "dev=%d", minor(dev));
	  ism_terminate(sp);
     }
     sdebug(2, "done");
     return 0;
}

static int
iscsi_ioctl(struct cdev *dev, u_long cmd, caddr_t arg, int mode, struct thread *td)
{
     struct isc		*sc;
     isc_session_t	*sp;
     isc_opt_t		*opt;
     int		error, fd;

     sc = (struct isc *)dev->si_drv1;
     sp = (isc_session_t *)dev->si_drv2;
     debug_called(8);

     error = 0;
     if(minor(dev) == MAX_SESSIONS) {
	  /*
	   | non Session commands
	   */
	  if(sc == NULL)
	       return ENXIO;

	  switch(cmd) {
	  case ISCSISETSES:
	       error = i_create_session(dev, (int *)arg);
	       break;

	  default:
	       error = ENXIO; // XXX: 
	  }
	  return error;
     }
     /*
      | session commands
      */
     if(sp == NULL)
	  return ENXIO;

     sdebug(6, "dev=%d cmd=%d", minor(dev), (int)(cmd & 0xff));

     switch(cmd) {
     case ISCSISETSOC:
	  fd = *(u_int *)arg;

	  if(sp->soc != NULL) {
	       isc_stop_receiver(sp);
	       fputsock(sp->soc);
	       sp->soc = NULL;
	  }
	  if((error = fgetsock(td, fd, &sp->soc, 0)) == 0)
	       isc_start_receiver(sp);
	  break;

     case ISCSISETOPT:
	  opt = (isc_opt_t *)arg;
	  error = i_setopt(sp, opt);
	  break;

     case ISCSISEND:
	  error = i_send(dev, arg, td);
	  break;

     case ISCSIRECV:
	  error = i_recv(dev, arg, td);
	  break;

     case ISCSIPING:
	  error = i_ping(dev);
	  break;

     case ISCSISTART:
	  error = ism_fullfeature(dev, 1);
	  break;

     case ISCSISTOP:
	  error = ism_fullfeature(dev, 0);
	  break;
	  
     case ISCSISIGNAL: {
	  int sig = *(int *)arg;

	  if(sig < 0 || sig > _SIG_MAXSIG)
	       error = EINVAL;
	  else {
		sp->signal = sig;
		sp->proc = sig ? td->td_proc : NULL;
	  }
	  break;
     }

     case ISCSIGETCAM: {
	  iscsi_cam_t *cp = (iscsi_cam_t *)arg;

	  error = ic_getCamVals(sp, cp);
	  break;
     }

     default:
	  error = ENOIOCTL;
     }

     return error;
}

#ifdef  ISCSI_DEBUG
static int
iscsi_read(struct cdev *dev, struct uio *uio, int ioflag)
{
     struct isc_softc	*sc;
     isc_session_t	*sp;
     pduq_t 		*pq;
     char		buf[1024];

     sc = (struct isc_softc *)dev->si_drv1;
     sp = (isc_session_t *)dev->si_drv2;
     if(minor(dev) == MAX_SESSIONS) {
	  sprintf(buf, "/----- Session ------/\n");
	  uiomove(buf, strlen(buf), uio);
	  int	i = 0;

	  TAILQ_FOREACH(sp, &sc->isc_sess, sp_link) {
	       if(uio->uio_resid == 0)
		    return 0;
	       sprintf(buf, "%03d] '%s' '%s'\n", i++, sp->opt.targetAddress, sp->opt.targetName);
	       uiomove(buf, strlen(buf), uio);
	  }
     }
     else {
	  int	i = 0;
	  struct socket	*so = sp->soc;
#define pukeit(i, pq) do {sprintf(buf, "%03d] %06x %02x %x\n", i, ntohl( pq->pdu.ipdu.bhs.CmdSN), \
	pq->pdu.ipdu.bhs.opcode, ntohl(pq->pdu.ipdu.bhs.itt));} while(0)

	  sprintf(buf, "/---- hld -----/\n");
	  uiomove(buf, strlen(buf), uio);
	  TAILQ_FOREACH(pq, &sp->hld, pq_link) {
	       if(uio->uio_resid == 0)
		    return 0;
	       pukeit(i, pq); i++;
	       uiomove(buf, strlen(buf), uio);
	  }
	  sprintf(buf, "/---- rsp -----/\n");
	  uiomove(buf, strlen(buf), uio);
	  i = 0;
	  TAILQ_FOREACH(pq, &sp->rsp, pq_link) {
	       if(uio->uio_resid == 0)
		    return 0;
	       pukeit(i, pq); i++;
	       uiomove(buf, strlen(buf), uio);
	  }
	  sprintf(buf, "/---- csnd -----/\n");
	  i = 0;
	  uiomove(buf, strlen(buf), uio);
	  TAILQ_FOREACH(pq, &sp->csnd, pq_link) {
	       if(uio->uio_resid == 0)
		    return 0;
	       pukeit(i, pq); i++;
	       uiomove(buf, strlen(buf), uio);
	  }
	  sprintf(buf, "/---- wsnd -----/\n");
	  i = 0;
	  uiomove(buf, strlen(buf), uio);
	  TAILQ_FOREACH(pq, &sp->wsnd, pq_link) {
	       if(uio->uio_resid == 0)
		    return 0;
	       pukeit(i, pq); i++;
	       uiomove(buf, strlen(buf), uio);
	  }
	  sprintf(buf, "/---- isnd -----/\n");
	  i = 0;
	  uiomove(buf, strlen(buf), uio);
	  TAILQ_FOREACH(pq, &sp->isnd, pq_link) {
	       if(uio->uio_resid == 0)
		    return 0;
	       pukeit(i, pq); i++;
	       uiomove(buf, strlen(buf), uio);
	  }

	  sprintf(buf, "/---- Stats ---/\n");
	  uiomove(buf, strlen(buf), uio);

	  sprintf(buf, "recv=%d sent=%d\n", sp->stats.nrecv, sp->stats.nsent);
	  uiomove(buf, strlen(buf), uio);

	  sprintf(buf, "flags=%x pdus: free=%d alloc=%d max=%d\n", 
		  sp->flags, sc->npdu_free, sc->npdu_alloc, sc->npdu_max);
	  uiomove(buf, strlen(buf), uio);

	  sprintf(buf, "cws=%d last cmd=%x exp=%x max=%x stat=%x itt=%x\n",
		  sp->cws, sp->sn.cmd, sp->sn.expCmd, sp->sn.maxCmd, sp->sn.stat, sp->sn.itt);
	  uiomove(buf, strlen(buf), uio);

	  sprintf(buf, "/---- socket -----/\nso_count=%d so_state=%x\n", so->so_count, so->so_state);
	  uiomove(buf, strlen(buf), uio);

     }
     return 0;
}
#endif

static int
i_ping(struct cdev *dev)
{
     return 0;
}
/*
 | low level I/O
 */
static int
i_send(struct cdev *dev, caddr_t arg, struct thread *td)
{
     isc_session_t	*sp = (isc_session_t *)dev->si_drv2;
     struct isc_softc	*sc = (struct isc_softc *)dev->si_drv1;
     caddr_t		bp;
     pduq_t		*pq;
     pdu_t		*pp;
     int		n, error;

     debug_called(8);

     if((pq = pdu_alloc(sc, 0)) == NULL)
	  return EAGAIN;
     pp = &pq->pdu;

     pq->pdu = *(pdu_t *)arg;
     if((error = i_prepPDU(sp, pq)) != 0)
	  return error;

     sdebug(3, "len=%d ahs_len=%d ds_len=%d", pq->len, pp->ahs_len, pp->ds_len);

     pq->buf = bp = malloc(pq->len - sizeof(union ipdu_u), M_ISCSI, M_WAITOK);

     if(pp->ahs_len) {
	  n = pp->ahs_len;
	  copyin(pp->ahs, bp, n);
	  pp->ahs = (ahs_t *)bp;
	  bp += n;
     }
     if(pp->ds_len) {
	  n = pp->ds_len;
	  copyin(pp->ds, bp, n); // can fail ...
	  pp->ds = bp;
	  bp += n;
	  while(n & 03) {
	       n++;
	       *bp++ = 0;
	  }
     }

     isc_qout(sp, pq);

     return error;
}

/*
 | NOTE: must calculate digest if requiered.
 */
static int
i_recv(struct cdev *dev, caddr_t arg, struct thread *td)
{
     isc_session_t	*sp = (isc_session_t *)dev->si_drv2;
     pduq_t		*pq;
     pdu_t		*pp, *up;
     caddr_t		bp;
     int		error;
     size_t		need, have, n;

     debug_called(8);

     if(sp == NULL)
	  return EIO;

     // XXX: maybe the user can request a time out?
     mtx_lock(&sp->rsp_mtx);
     while((pq = i_dqueue_rsp(sp)) == NULL && !(sp->flags & ISC_CON_STOPPED)) {
	  msleep(&sp->rsp, &sp->rsp_mtx, PRIBIO, "isc_rsp", hz*20);
     }
     mtx_unlock(&sp->rsp_mtx);

     if(pq == NULL) {
	  error = ENOTCONN;
	  sdebug(3, "error=%d sp->flags=%x ", error, sp->flags);
	  return error;
     }
     up = (pdu_t *)arg;
     pp = &pq->pdu;
     up->ipdu = pp->ipdu;
     n = 0;
     up->ds_len = 0;
     up->ahs_len = 0;
     error = 0;

     if(pq->mp) {
	  bp = mtod(pq->mp, caddr_t);
	  /*
	   | NOTE: this functions does NOT KNOW (yet) how to deal with
	   | a mbuf data that is less than len
	   | if(mp->m_len < len) ... is this a 'cannot happen case'?
	   */
	  if(pp->ahs_len) {
	       need = pp->ahs_len;
	       if(sp->hdrDigest)
		    need += 4;
	       n = MIN(up->ahs_size, need);
	       error = copyout(bp, (caddr_t)up->ahs, n);
	       up->ahs_len = n;
	       bp += need;
	  }
	  if(!error && pp->ds_len) {
	       need = pp->ds_len;
	       if(sp->hdrDigest)
		    need += 4;
	       if((have = up->ds_size) == 0) {
		    have = up->ahs_size - n;
		    up->ds = (caddr_t)up->ahs + n;
	       }
	       n = MIN(have, need);
	       error = copyout(bp, (caddr_t)up->ds, n);
	       up->ds_len = n;
	  }
     }

     sdebug(3, "len=%d ahs_len=%d ds_len=%d", pq->len, pp->ahs_len, pp->ds_len);

     pdu_free(sp->isc, pq);

     return error;
}

static int
i_create_session(struct cdev *dev, int *ndev)
{ 
     struct isc_softc		*sc = (struct isc_softc *)dev->si_drv1;
     isc_session_t	*sp;
     int		error, n;

     debug_called(8);
     sp = (isc_session_t *)malloc(sizeof *sp, M_ISCSI, M_WAITOK | M_ZERO);
     if(sp == NULL)
	  return ENOMEM;

     mtx_lock(&sc->mtx);
     /*
      | search for the lowest unused sid
      */
     for(n = 0; n < MAX_SESSIONS; n++)
	  if(sc->sessions[n] == NULL)
	       break;
     if(n == MAX_SESSIONS) {
	  mtx_unlock(&sc->mtx);
	  return EPERM;
     }
     TAILQ_INSERT_TAIL(&sc->isc_sess, sp, sp_link);
     sc->nsess++;
     mtx_unlock(&sc->mtx);

     sc->sessions[n] = sp;
     sp->dev = make_dev(&iscsi_cdevsw, n, UID_ROOT, GID_WHEEL, 0600, "iscsi%d", n);
     *ndev = sp->sid = n;
     sp->isc = sc;
     sp->dev->si_drv1 = sc;
     sp->dev->si_drv2 = sp;

     sp->opt.maxRecvDataSegmentLength = 8192;
     sp->opt.maxBurstLength = 65536;	// 64k

     sdebug(3, "sessionID=%d", n);

     error = ism_start(sp);

     return error;
}

static void
iscsi_start(void)
{
     struct isc_softc *sc = &isc;

     debug_called(8);

     memset(sc, 0, sizeof(struct isc_softc));

     sc->dev = make_dev(&iscsi_cdevsw, MAX_SESSIONS, UID_ROOT, GID_WHEEL, 0600, "iscsi");
     sc->dev->si_drv1 = sc;

     TAILQ_INIT(&sc->isc_sess);
     
     sc->pdu_zone = uma_zcreate("pdu", sizeof(pduq_t),
				NULL, NULL, NULL, NULL,
				0, 0);
     if(sc->pdu_zone == NULL) {
	  printf("iscsi_initiator: uma_zcreate failed");
	  // XXX: and now what?
     }
     uma_zone_set_max(sc->pdu_zone, MAX_PDUS*2+1);
#ifdef ISCSI_DEBUG
     sc->npdu_free = MAX_PDUS*2+1;
#endif
     mtx_init(&sc->mtx, "iscsi", NULL, MTX_DEF);
     mtx_init(&sc->mtx_pdu, "iscsi pdu pool", NULL, MTX_DEF);
#if 0
     // XXX: this will cause a panic if the
     //      module is loaded too early
     if(ic_init(sc) != 0)
	  return;
#else
     sc->cam_sim = NULL;
#endif
     debug(3, "done");
}

/*
 | Notes:
 |	unload SHOULD fail if there is activity
 |	activity: there is/are active session/s
 */
static void
iscsi_stop(void)
{
     struct isc_softc	*sc = &isc;
     isc_session_t	*sp, *sp_tmp;

     debug_called(8);

     /*
      | got through all the sessions
      | Note: close should have done this ...
      */
     TAILQ_FOREACH_SAFE(sp, &sc->isc_sess, sp_link, sp_tmp) {
	  //XXX: check for activity ...
	  ism_terminate(sp);
     }

     ic_destroy(sc);

     uma_zdestroy(sc->pdu_zone);

     if(sc->dev)
	  destroy_dev(sc->dev);
}

static int
iscsi_modevent(module_t mod, int what, void *arg)
{
     debug_called(8);

     switch(what) {
     case MOD_LOAD:
	  iscsi_start();
	  break;

     case MOD_QUIESCE:
#if notyet
	  if(isc.nsess)
	       log(LOG_ERR, "iscsi module busy, cannot unload");
	  return isc.nsess;
#endif
     case MOD_SHUTDOWN:
	  break;

     case MOD_UNLOAD:
	  iscsi_stop();
	  break;

     default:
	  break;
     }
     return 0;
}
moduledata_t iscsi_mod = {
         "iscsi",
         (modeventhand_t) iscsi_modevent,
         0
};

DECLARE_MODULE(iscsi, iscsi_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);
